package edu.ntua.ivml.qa.algorithm;

import java.io.PrintStream;
import java.util.ArrayList;

import edu.ntua.ivml.qa.db.QueryCollector;
import edu.ntua.ivml.qa.lp.Clause;


public class Executor {

	private Engine engine;
	
	public int allRewritingsSize;
	public int coreRewritingsSize;
	public long reasoningTime;
	public long redundantRemovalTime;
	
	public Executor(Engine engine) throws Exception {
		this.engine = engine;
	}
	
	public ExecutionResult threadExecute(Clause query, PrintStream out) throws Exception {

		QueryCollector qc = engine.createQueryCollector(query);
	
		engine.start();
	
		while (engine.isAlive()) {
			engine.join(100);
		}
		
		System.out.println("Rewriting computation finished.");

		ArrayList<Clause> rewritings = qc.getQueries();

		long sTime = System.currentTimeMillis();  
		
		ArrayList<Clause> frewritings = qc.getNonRedundantQueries();
		
		System.out.println("Redundant rewritings removal finished.");
		
		long fTime = System.currentTimeMillis() - sTime;

		allRewritingsSize = rewritings.size();
		coreRewritingsSize = frewritings.size();
		
		reasoningTime = engine.getLastExecutionTime();
		redundantRemovalTime = fTime;
		
		return new ExecutionResult(frewritings);
	}

}


