package edu.ntua.ivml.qa.algorithm.rapid;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import edu.ntua.ivml.qa.dllite.DLLiteRTheory;
import edu.ntua.ivml.qa.lp.Atom;
import edu.ntua.ivml.qa.lp.Clause;
import edu.ntua.ivml.qa.lp.Function;
import edu.ntua.ivml.qa.lp.FunctionalTerm;
import edu.ntua.ivml.qa.lp.Predicate;
import edu.ntua.ivml.qa.lp.Substitution;
import edu.ntua.ivml.qa.lp.Term;
import edu.ntua.ivml.qa.lp.Variable;
import edu.ntua.ivml.qa.utils.Utils;

public class AtomUnfolder {

	private DLLiteRTheory program;
	private Map<Atom, AtomUnfoldList> map;
	
	public AtomUnfolder(DLLiteRTheory program) {
		this.program = program;

		map = new HashMap<Atom, AtomUnfoldList>();
	}

	public Map<Atom, AtomUnfoldList> getAtomMap() {
		return map;
	}

	public ArrayList<UnfoldedClause> unfold(UnfoldedClause cc) {
		return unfold(cc, -1, null);
	}
	
	public ArrayList<UnfoldedClause> unfold(UnfoldedClause cc, int eIndex, Set<Atom> extraAtoms) {
		Clause clause = cc.getClause();
		
		int[] index = new int[clause.bodySize()];
		index[0] = -1;
		
		Variable[][] uVarMap = new Variable[clause.bodySize()][2];

		for (int i = 0; i < clause.bodySize(); i++) {
			Term[] terms = clause.getBodyAtomAt(i).getArguments().getTerms();
			for (int j = 0; j < terms.length; j++) {
				if (terms[j].hasUnboundPrefix()) {
					uVarMap[i][j] = (Variable)terms[j];
				}
			}
		}		
		
		Substitution subst = clause.replaceUnboundVariablesByUnderscoreSubstitution();
		Clause uClause = clause.apply(subst);
		
		Set<Atom> uExtraAtoms = null;
		if (extraAtoms != null) {
			uExtraAtoms = new HashSet<Atom>();
			for (Atom a : extraAtoms) {
				uExtraAtoms.add(a.apply(subst));
			}
		}
			
		ArrayList<UnfoldedClause> result = new ArrayList<UnfoldedClause>();
		
		unfold(uClause.getHead(), uVarMap, createAtomList(uClause.getBody(), eIndex, uExtraAtoms), 
			   index, 0,  new HashSet<Integer>(), result, 
			   new ArrayList<ArrayList<Integer>>());
		

		return result;
	}

	public Clause testUnfold(Clause clause) {
		
		int[] index = new int[clause.bodySize()];
		index[0] = -1;
		
		Variable[][] uVarMap = new Variable[clause.bodySize()][2];

		for (int i = 0; i < clause.bodySize(); i++) {
			Term[] terms = clause.getBodyAtomAt(i).getArguments().getTerms();
			for (int j = 0; j < terms.length; j++) {
				if (terms[j].hasUnboundPrefix()) {
					uVarMap[i][j] = (Variable)terms[j];
				}
			}
		}		
		
		Substitution subst = clause.replaceUnboundVariablesByUnderscoreSubstitution();		
		Clause uClause = clause.apply(subst);
		
		ArrayList<AtomIndexEntryList> aieList = createAtomList(uClause.getBody());

		Set<Integer> remove = new HashSet<Integer>();
		
		ArrayList<Atom> newClause = new ArrayList<Atom>(); 
		for (int i = 0; i < aieList.size(); i++) {
			int first = aieList.get(i).first;
			
			ArrayList<Integer> cover = aieList.get(i).atomList.get(first).covered; 
			for (int k : cover) {
				if (k != i) {
					remove.add(k);
				}
			}
		}
		
		for (int i = 0; i < clause.getBody().size(); i++) {
			if (!remove.contains(i)) {
				newClause.add(clause.getBodyAtomAt(i));
			}
			
		}
		
		return new Clause(clause.getHead(), newClause);
	}

	public static int mode = 1;
	
	private boolean isCoveringOK(ArrayList<ArrayList<Integer>> history) {
		Set<Integer> common = new HashSet<Integer>();
		Set<Integer> nonCommon = new HashSet<Integer>();

		for (ArrayList<Integer> list : history) {
			for (int i : list) {
				if (common.contains(i)) {
					
				} else if (nonCommon.contains(i)) {
					common.add(i);
					nonCommon.remove(i);
				} else {
					nonCommon.add(i);
				}
			}
		}
		
		for (ArrayList<Integer> list : history) {
			if (common.containsAll(list)) {
				return false;
			}
		}
		
		return true;
	}
	
	private Clause createBodyFromList(Atom head, Variable[][] uVarMap, ArrayList<AtomIndexEntryList> atomList, int[] index) {
		ArrayList<Atom> newList = new ArrayList<Atom>();
		
		int count = 0;
		
		for (int j = 0; j < atomList.size(); j++) {
			if (index[j] != -1) {
				AtomIndexEntry aie = atomList.get(j).atomList.get(index[j]);
				
				Term[] terms = aie.atom.getArguments().getTerms();
				
				Substitution s = new Substitution();
				
				for (int k = 0 ; k < terms.length ; k++) {
					if (terms[k].equals(Variable.UNBOUND_VARIABLE_1)) {
						s.add(Variable.UNBOUND_VARIABLE_1, uVarMap[j][0]);
					} else if (terms[k].equals(Variable.UNBOUND_VARIABLE_2)) {
						s.add(Variable.UNBOUND_VARIABLE_2, uVarMap[j][1]);
					} else if (terms[k].equals(RapidStatics.OTHER_UNBOUND_VARIABLE_1)) {
						s.add(RapidStatics.OTHER_UNBOUND_VARIABLE_1, new Variable(Variable.UNBOUND_NEW_PREFIX + count++));
					} else if (terms[k].equals(RapidStatics.OTHER_UNBOUND_VARIABLE_2)) {
						s.add(RapidStatics.OTHER_UNBOUND_VARIABLE_2, new Variable(Variable.UNBOUND_NEW_PREFIX + count++));
					}
				}
				
				newList.add(aie.atom.apply(s));
			}
		}
		
		return new Clause(head, newList);

	}
	
	private void unfold(Atom head, Variable[][] uVarMap, ArrayList<AtomIndexEntryList> atomList, int[] index, int k, Set<Integer> next, ArrayList<UnfoldedClause> result, ArrayList<ArrayList<Integer>> history) {
		
		if (k == atomList.size()) {
			
			if (!isCoveringOK(history)) {
				return;
			}
			
			boolean pure = true;
			
			Set<AtomIndexEntry> aies = new HashSet<AtomIndexEntry>();
			Set<AtomIndexEntry> roots = new HashSet<AtomIndexEntry>();
			Set<AtomIndexEntry> childrenRoots = new HashSet<AtomIndexEntry>();
			
			for (int j = 0; j < k; j++) {
				if (index[j] != -1) {
					AtomIndexEntry aie = atomList.get(j).atomList.get(index[j]);
					
					if (aie.roots != null) {
						pure = false;
					} 
					
					if (mode == 2) {
						aies.add(aie);
						if (aie.roots != null) {
							roots.addAll(aie.roots);
						}
						
						if (aie.children != null) {
							childrenRoots.add(aie);
						}
					} 
				}
			}
			
			Clause newClause = null;

			if (mode == 2) {
				if (!roots.isEmpty()) {
					AtomIndexEntry[] rootsArray = roots.toArray(new AtomIndexEntry[] {});
					
					for (int i = 0; i < rootsArray.length; i++) {
						Set<AtomIndexEntry> tmp = new HashSet<AtomIndexEntry>();
						tmp.add(rootsArray[i]);
						
						for (AtomIndexEntry a : aies) {
							if (!rootsArray[i].children.contains(a)) {
								tmp.add(a);
							}
						}

						Set<Integer> covered = new HashSet<Integer>();
						ArrayList<ArrayList<Integer>> tmpHistory = new ArrayList<ArrayList<Integer>>(); 
						for (AtomIndexEntry a : tmp) {
							tmpHistory.add(a.covered);
							covered.addAll(a.covered);
						}
	
						if (covered.size() == atomList.size() && isCoveringOK(tmpHistory)) {
							return;
						}
					}
				}
				
				if (!childrenRoots.isEmpty()) {
					AtomIndexEntry[] rootsArray = childrenRoots.toArray(new AtomIndexEntry[] {});

					Map<Atom, ArrayList<Integer>> childrenAtoms = new HashMap<Atom, ArrayList<Integer>>();
					
					for (int i = 0; i < rootsArray.length; i++) {
						for (AtomIndexEntry a : rootsArray[i].children) {
							childrenAtoms.put(a.atom, a.covered);
						}
					}
					
					for (Map.Entry<Atom, ArrayList<Integer>> entry : childrenAtoms.entrySet()) {
						Atom atom = entry.getKey();
						ArrayList<Integer> atomCover = entry.getValue();

						Set<Atom> tmp = new HashSet<Atom>();
						tmp.add(atom);
						
						for (AtomIndexEntry aie : aies) {
							if (!atomCover.containsAll(aie.covered)) {
								tmp.add(aie.atom);
							}
						}
					
						newClause = createBodyFromList(head, uVarMap, atomList, index);
							
						if (newClause.isSubsumedBy(new Clause(head, tmp).condense())) {
							return;
						}
					}
				}
			}
			
			if (newClause == null) {
				newClause = createBodyFromList(head, uVarMap, atomList, index);
			}

			result.add(new UnfoldedClause(newClause, pure));
			
		} else {
			if (next.contains(k)) {
				Arrays.fill(index, k, index.length, -1);
				
				unfold(head, uVarMap, atomList, index, k + 1, next, result, history);
			} else {
				loop:
				for (int i = 0; i < atomList.get(k).atomList.size(); i++) {
					index[k]++;
					Arrays.fill(index, k + 1, index.length, -1);
					
					AtomIndexEntry currentEntry = atomList.get(k).atomList.get(i);
					
					Set<Integer> newNext = new HashSet<Integer>();
					newNext.addAll(next);
					if (currentEntry.fIndex != null) {
						newNext.addAll(currentEntry.fIndex);
					}

					ArrayList<ArrayList<Integer>> newHistory = new ArrayList<ArrayList<Integer>>(); 
					if (Utils.subsumesSome(history, currentEntry.covered)) {
						continue loop;
					} else {
						ArrayList<ArrayList<Integer>> check = new ArrayList<ArrayList<Integer>>();

						int top = currentEntry.covered.get(0);
						for (ArrayList<Integer> list : history) {
							if (list.get(0) == top) {
								check.add(list);
							}
						}

						ArrayList<ArrayList<Integer>> res = new ArrayList<ArrayList<Integer>>();
						
						int it = 1;
						inloop:
						while (check.size() > 0 && currentEntry.covered.size() > it) {
							int itop = currentEntry.covered.get(it);
							for (ArrayList<Integer> list : check) {
								if (list.size() > it) {
									if (list.get(it) == itop) {
										res.add(list);
									} else if (list.get(it) > itop) {
										continue loop;
									} else {
										break inloop;
									}
								}
							}
							it++;
								
							check.addAll(res);
							res.clear();
						}
						
						newHistory.addAll(history);
						newHistory.add(currentEntry.covered);
					}
					
					unfold(head, uVarMap, atomList, index, k + 1, newNext, result, newHistory);
				}
			}
		}
	}
	
	// atom should have unbound variables as ?_ only.
	public AtomUnfoldResult getUnfolding(Atom atom) {
		ProperRenamedAtom pra = new ProperRenamedAtom(atom); 
		
		Atom searchAtom = pra.getRenamedAtom();
		AtomUnfoldList searchList = map.get(searchAtom);
		
		if (searchList == null) {
			searchList = unfold(searchAtom);
			map.put(searchAtom, searchList);
		}		
		
		return new AtomUnfoldResult(searchList, pra.getSubstitution());
	}


	public AtomUnfoldResult getUnfolding(Set<Atom> atoms) {
		
		ArrayList<Atom> all = new ArrayList<Atom>();
		
		Substitution subst = new Substitution();

		int i = 0;
		for (Atom atom : atoms) {
			ProperRenamedAtom pra = new ProperRenamedAtom(atom); 
			
			if (!subst.add(pra.getSubstitution())) {
				System.out.println("VARIABLE ERROR");
				return null;
			}
			
			Atom searchAtom = pra.getRenamedAtom();
			AtomUnfoldList searchList = map.get(searchAtom);
		
			if (searchList == null) {
				searchList = unfold(searchAtom);
				map.put(searchAtom, searchList);
			}
			
			if (i == 0) {
				for (Atom a : searchList.list) {
					all.add(a);
				}
			} else {
				loop:
				for (Atom a : searchList.list) {
					int k = all.size();
					for (int j = 0; j < k; j++) {
        				Atom ex = all.get(j);
        				if (ex.equals(a)) {
        					continue loop;
        				} else if (ex.unboundEquals(a)) {
//            				if (!ex.getVariables().contains(Variable.UNBOUND_VARIABLE)) {
//            					all.remove(j);
//            					all.add(a);
//            				}
//            				continue loop;
        					
        					int exU = 0;
        					for (Variable v : ex.getVariables()) {
        						if (v.getName().startsWith(RapidStatics.OTHER_UNBOUND_PREFIX)) {
        							exU++;
        						}
        					}

        					int aU = 0;
        					for (Variable v : a.getVariables()) {
        						if (v.getName().startsWith(RapidStatics.OTHER_UNBOUND_PREFIX)) {
        							aU++;
        						}
        					}

            				if (exU > aU) {
            					all.remove(j);
            					all.add(a);
            				}
            				
            				continue loop;
        				} 
        			}
					all.add(a);
				}
			}
			i++;
		}
		
    	Atom[] arr = all.toArray(new Atom[] {});

    	Arrays.sort(arr, new Comparator<Atom>() {
			public int compare(Atom a1, Atom a2) {
				return a1.getPredicate().compareTo(a2.getPredicate());
			} 
    	});

		return new AtomUnfoldResult(new AtomUnfoldList(arr), subst);
		
	}
	
	private class AtomIndexEntryList {
		public ArrayList<AtomIndexEntry> atomList;
		public int first;
		
		public AtomIndexEntryList(ArrayList<AtomIndexEntry> atomList, int first) {
			this.atomList = atomList;
			this.first = first;
		}
	}
	
	private ArrayList<AtomIndexEntryList> createAtomList(ArrayList<Atom> topAtoms) {
		return createAtomList(topAtoms, -1, null);
	}
	
	private ArrayList<AtomIndexEntryList> createAtomList(ArrayList<Atom> topAtoms, int eIndex, Set<Atom> extraAtoms) {
		ArrayList<Atom[]> aList = new ArrayList<Atom[]>();
		Substitution[] aSubst = new Substitution[topAtoms.size()];
		
		Set<Atom> allAtoms = null;
		
		for (int k = 0; k < topAtoms.size(); k++) {
			AtomUnfoldResult aur;
			
			if (k != eIndex) {
				aur = getUnfolding(topAtoms.get(k));
			} else {
				allAtoms = new HashSet<Atom>();
				allAtoms.addAll(extraAtoms);
				allAtoms.add(topAtoms.get(k));
				aur = getUnfolding(allAtoms);
			}

			if (!aur.substitution.isEmpty()) {
				aSubst[k] = aur.substitution;
			}
			
			aList.add(aur.unfoldList.list);

		}

		ArrayList<AtomIndexEntryList> atomList = new ArrayList<AtomIndexEntryList>();
		
		for (int k = 0; k < topAtoms.size(); k++) {
			ArrayList<AtomIndexEntry> currentList = new ArrayList<AtomIndexEntry>();
			
			Atom[] atoms = aList.get(k); 
			for (Atom a : atoms) {
				AtomIndexEntry aie;
				if (aSubst[k] == null) {
					aie = new AtomIndexEntry(a, k);
				} else {
					aie = new AtomIndexEntry(a.apply(aSubst[k]), k);
				}

				currentList.add(aie);
			}

			atomList.add(new AtomIndexEntryList(currentList, -1));

			for (int i = 0; i < currentList.size(); i++) {
				exactLinkBack(atomList, k, i);
				subsumeLinkBack(atomList, k, i);				
			}
			
		}
		
		for (int k = 0; k < topAtoms.size(); k++) {
			Atom atom = topAtoms.get(k);
		
			ArrayList<AtomIndexEntry> currentList = atomList.get(k).atomList;
			
			Predicate aPred = atom.getPredicate();
			int pos = binarySearch(currentList, aPred);
    	
			while (currentList.get(pos).atom.getPredicate().equals(aPred)) {
				if (currentList.get(pos).atom.unboundEquals(atom)) {
					atomList.get(k).first = pos;
					break;
				} else {
					pos++;
				}
			}
		}

		
//		System.out.println("----- " + topAtoms);
//		for (int i = 0; i < atomList.size(); i++) {
//			ArrayList<AtomIndexEntry> sList = atomList.get(i).atomList;
//			
//			for (int j = 0; j < sList.size(); j++) {
//				AtomIndexEntry aie = sList.get(j);
//				System.out.println(aie.pos + (atomList.get(i).first == j?"* : ":"  : ") + aie.atom +  " * " + aie.fIndex + " ***< " + aie.mainRoot + " " + (aie.covered) + " --- " + aie.roots + " " + ((aie.roots != null) ? aie.roots.iterator().next().children : ""));
//			}
//			System.out.println();
//		}

		return atomList;
	}
	
	private void exactLinkBack(ArrayList<AtomIndexEntryList> atomList, int k, int i) {
		AtomIndexEntry currentEntry = atomList.get(k).atomList.get(i);
		
		for (int j = k - 1; j >= 0; j--) {
			ArrayList<AtomIndexEntry> prevList = atomList.get(j).atomList;
			
			int p = exactAtomFind(j, prevList, currentEntry.atom);
			if (p != -1) {
				propagateBack(prevList.get(p), currentEntry, k); 

				return;
			} 
		}
		
		currentEntry.covered = new ArrayList<Integer>();
		currentEntry.covered.add(k);

	}
	
	private void subsumeLinkBack(ArrayList<AtomIndexEntryList> atomList, int k, int i) {
		ArrayList<AtomIndexEntry> currentList = atomList.get(k).atomList;
		AtomIndexEntry currentEntry = currentList.get(i);
		
		for (int j = k - 1; j >= 0; j--) {
			ArrayList<AtomIndexEntry> prevList = atomList.get(j).atomList;
			
			int p = subsumeAtomFind(j, prevList, currentEntry.atom);
			if (p != -1) {
				if (insType == 2) {
					int si = lastInsertIfNotInList(currentList, i, insAtom);
					if (si != -1) {
						AtomIndexEntry newEntry = new AtomIndexEntry(insAtom, k);
						newEntry.mainRoot = currentEntry;
						
						currentList.add(si, newEntry);
					}
						
				} else if (insType == 1) {
					int si = lastInsertIfNotInList(prevList, p, insAtom);
					if (si != -1) {
						AtomIndexEntry newEntry = new AtomIndexEntry(insAtom, j);
						newEntry.mainRoot = prevList.get(p);
						
						prevList.add(si, newEntry);
						
						exactLinkBack(atomList, j, si);
						subsumeLinkBack(atomList, j, si);
						
						propagateBack(newEntry, currentEntry, k);
												
						return;
					}
				} 
			}
		}
	}
	
	private int lastInsertIfNotInList(ArrayList<AtomIndexEntry> list, int i, Atom atom) {
		Predicate pred = atom.getPredicate();
	
		int fI = i;
		while (fI >= 0 && list.get(fI).atom.getPredicate().equals(pred)) {
			if (list.get(fI).atom.unboundEquals(insAtom)) {
				return -1;
			}
			fI--;
		}
		
		int sI = i + 1;
		while (sI < list.size() && list.get(sI).atom.getPredicate().equals(pred)) {
			if (list.get(sI).atom.unboundEquals(insAtom)) {
				return -1;
			}
			sI++;
		}
		
		return sI;
	}

	private void propagateBack(AtomIndexEntry matchedEntry, AtomIndexEntry newEntry, int k) {
		matchedEntry.nextEntry = newEntry; 
		
		AtomIndexEntry firstEntry = matchedEntry;
		if (matchedEntry.firstEntry != null) {
			firstEntry = matchedEntry.firstEntry;
		}
		newEntry.firstEntry = firstEntry;

		Set<AtomIndexEntry> roots = new HashSet<AtomIndexEntry>();
		if (newEntry.mainRoot != null) {
			roots.add(newEntry.mainRoot);
		}

		AtomIndexEntry entry = firstEntry;
		do {
			if (entry.fIndex == null) {
				entry.fIndex = new ArrayList<Integer>();
			}
			entry.fIndex.add(k);
			
			if (entry.roots != null) {
				roots.addAll(entry.roots);
			}
			if (entry.mainRoot != null) {
				roots.add(entry.mainRoot);
			}
			
			entry = entry.nextEntry;
		} while (entry != newEntry);

		if (roots.size() > 0) {
			Set<AtomIndexEntry> allChildren = new HashSet<AtomIndexEntry>();

			for (AtomIndexEntry root : roots) {
				if (root.children != null) {
					for (AtomIndexEntry ch : root.children) {
						allChildren.add(ch);
					}
				}
			}
			
			entry = firstEntry;
			
			do {
				entry.roots = roots;
				allChildren.add(entry);
				
				entry = entry.nextEntry;
			} while (entry != null);
			
			
			for (AtomIndexEntry cRoot : roots) {
				Set<AtomIndexEntry> children = new HashSet<AtomIndexEntry>();
				for (AtomIndexEntry c : allChildren) {
					atomSubsumes(cRoot.atom, c.atom);
					if (subsumptionMatchType == 1) {
						children.add(c);
					}
				}
				cRoot.children = children;
			}
			
			entry = firstEntry;
		}
		
		newEntry.covered = matchedEntry.covered;
		newEntry.covered.add(k);
	}
	
		
	private class AtomIndexEntry {
		Atom atom;
		int pos;
		ArrayList<Integer> fIndex;
		AtomIndexEntry firstEntry;
		AtomIndexEntry nextEntry;
		
		ArrayList<Integer> covered;
		
		AtomIndexEntry mainRoot;
		
		Set<AtomIndexEntry> roots;
		Set<AtomIndexEntry> children;
		
		public AtomIndexEntry(Atom atom, int pos) {
			this.atom = atom;
			this.pos = pos;
		}
		
		public String toString() {
			return atom.toString();
		}
	}

	private int insType;
	private Atom insAtom;
	
	private int subsumeAtomFind(int c, ArrayList<AtomIndexEntry> list, Atom atom) {
		Predicate pred = atom.getPredicate();
		
		int pos = binarySearch(list, pred);
		
		insAtom = null;
		insType = -1;
		
		if (pos < 0) {
			return -1;
		} else {
			while (pos < list.size() && list.get(pos).atom.getPredicate().equals(pred)) {
				Atom mAtom = atomSubsumes(list.get(pos).atom, atom);
				
				if (subsumptionMatchType == 1) {
					insAtom = mAtom;					
					insType = 1;

					return pos;
					
				} else if (subsumptionMatchType == 2) {
					insAtom = mAtom;
					insType = 2;
					
					return pos;
					
				} else {
					pos++;
				}
			}
			
			return -1;
		}
	}	
	
	private int exactAtomFind(int c, ArrayList<AtomIndexEntry> list, Atom atom) {
		Predicate pred = atom.getPredicate();
		
		int pos = binarySearch(list, pred);
		
		if (pos < 0) {
			return -1;
		} else {
			while (pos < list.size() && list.get(pos).atom.getPredicate().equals(pred)) {
				if (list.get(pos).atom.getArguments().unboundEquals(atom.getArguments())) {
					return pos;
				}
				pos++;
			}
		}
		
		return -1;
	}
	
	private int subsumptionMatchType;
	
	private Atom atomSubsumes(Atom a1, Atom a2) {
		subsumptionMatchType = -1;
		
		if (a1.getArguments().unboundEquals(a2.getArguments())) {
			subsumptionMatchType = 0;
			return a1;
		}

		Substitution subst = new Substitution();
		
		int size = a1.getArguments().size();
		
		for (int i = 0; i < size; i++) {
			Term t1 = a1.getArguments().getTerm(i);
			Term t2 = a2.getArguments().getTerm(i);
			
			boolean t1u = t1.hasUnboundPrefix();
			boolean t2u = t2.hasUnboundPrefix();

			if (t1u && !t2u) {
				if (subsumptionMatchType == 2) {
					subsumptionMatchType = -1;
					return null;
				}
				subsumptionMatchType = 1;
				subst.add((Variable)t1, t2);
			} else if (!t1u && t2u) {
				if (subsumptionMatchType == 1) {
					subsumptionMatchType = -1;
					return null;
				}
				subsumptionMatchType = 2;
				subst.add((Variable)t2, t1);
			} else if (!t1u && !t2u && !t1.equals(t2)) {
				subsumptionMatchType = -1;
				return null;
			}
		}

		if (subsumptionMatchType == 1) {
			return a1.apply(subst);
		} else if (subsumptionMatchType == 2) {
			return a2.apply(subst);
		} else {
			return null;
		}
	}	

    private AtomUnfoldList unfold(Atom atom) {
    	
    	Set<Function> arg0Functions = new HashSet<Function>();
    	Set<Function> arg1Functions = new HashSet<Function>();
    	
    	ArrayList<Atom> result = new ArrayList<Atom>();
    	result.add(atom);   
    	
    	int i = 0;
    	while (i < result.size()) {
    		loop:
    		for (Atom a : generateResolvents(result.get(i++), arg0Functions, arg1Functions)) {
    			for (int j = 0; j < result.size(); j++) {
    				Atom ex = result.get(j);
    				if (ex.equals(a)) {
    					continue loop;
    				} else if (ex.unboundEquals(a)) {
    					int exU = 0;
    					for (Variable v : ex.getVariables()) {
    						if (v.getName().startsWith(RapidStatics.OTHER_UNBOUND_PREFIX)) {
    							exU++;
    						}
    					}

    					int aU = 0;
    					for (Variable v : a.getVariables()) {
    						if (v.getName().startsWith(RapidStatics.OTHER_UNBOUND_PREFIX)) {
    							aU++;
    						}
    					}

        				if (exU > aU) {
        					result.remove(j);
        					result.add(a);
        				}
        				
        				continue loop;
    				} 
    			}
    			
   				result.add(a);
    		}
    	}

    	Atom[] arr = result.toArray(new Atom[] {});

    	Arrays.sort(arr, new Comparator<Atom>() {
			public int compare(Atom a1, Atom a2) {
				return a1.getPredicate().compareTo(a2.getPredicate());
			} 
    	});
     	
    
    	return new AtomUnfoldList(arr, arg0Functions, arg1Functions); 
    }    	
    
    private ArrayList<Atom> generateResolvents(Atom atom, Set<Function> arg0Functions, Set<Function> arg1Functions) {
    	ArrayList<Atom> result = new ArrayList<Atom>();
    	
   		for (Clause sideClause : program.getClauses(atom.getPredicate())) {

			if (!sideClause.isFunctionFree()) {
				Term[] sideTuple = sideClause.getHead().getArguments().getTerms();

   				Term[] atomTuple = atom.getArguments().getTerms();
   				Term t0 = atomTuple[0];
				
				if (atom.isRole()) {
	   				Term t1 = atomTuple[1];

	   				if (sideTuple[0].isFunctionalTerm() && !t0.hasUnboundPrefix()) {
	   					if (t0.equals(RapidStatics.dummyVars.get(0))) {
	   						arg0Functions.add(((FunctionalTerm)sideTuple[0]).getFunction());
	   					} else {
	   						arg1Functions.add(((FunctionalTerm)sideTuple[0]).getFunction());
	   					}
	   					continue;
	   				} else if (sideTuple[1].isFunctionalTerm() && !t1.hasUnboundPrefix()) {
	   					if (t1.equals(RapidStatics.dummyVars.get(1))) {
	   						arg1Functions.add(((FunctionalTerm)sideTuple[1]).getFunction());
	   					} else {
	   						arg0Functions.add(((FunctionalTerm)sideTuple[1]).getFunction());
	   					}
	   					continue;
	   				}
				} else {
					if (!t0.hasUnboundPrefix()) {
						if (t0.equals(RapidStatics.dummyVars.get(0))) {
		   					arg0Functions.add(((FunctionalTerm)sideTuple[0]).getFunction());
		   				} else {
		   					arg1Functions.add(((FunctionalTerm)sideTuple[0]).getFunction());
		   				}
						continue;
					}
				}
  			} 
			
			Substitution s = Substitution.mgu(atom, sideClause.getHead());
			
			if (s != null) {
				
				Atom bodyAtom = sideClause.getBodyAtomAt(0); 
				if (bodyAtom.getVariables().contains(Variable.UNBOUND_VARIABLE)) {
					Set<Variable> sVars = bodyAtom.apply(s).getVariables();
					
					if (sVars.contains(RapidStatics.OTHER_UNBOUND_VARIABLE_1)) {
						s.add(Variable.UNBOUND_VARIABLE, RapidStatics.OTHER_UNBOUND_VARIABLE_2);
					} else {
						s.add(Variable.UNBOUND_VARIABLE, RapidStatics.OTHER_UNBOUND_VARIABLE_1);
					}
				}
				
		    	Atom newAtom = bodyAtom.apply(s);
		    	
		    	Set<Variable> vars = newAtom.getVariables();
		    	
		    	Substitution s2 = new Substitution();
		    	
		    	for (Variable v : vars) {
		    		if (!v.hasUnboundPrefix() && !RapidStatics.dummyVars.contains(v)) {
		    			if (vars.contains(RapidStatics.OTHER_UNBOUND_VARIABLE_1)) {
							s2.add(v, RapidStatics.OTHER_UNBOUND_VARIABLE_2);
						} else {
							s2.add(v, RapidStatics.OTHER_UNBOUND_VARIABLE_1);
						}
		    		}
		    	}
		    	
				result.add(newAtom.apply(s2));
			}
   		}
        
        return result;
    } 
	
    private static int binarySearch(ArrayList<AtomIndexEntry> a, Predicate key) {
		int low = 0;
		int high = a.size() - 1;
		
		while (low <= high) {
			int mid = (low + high) >>> 1;
			AtomIndexEntry midVal = a.get(mid);
			
			int cmp = midVal.atom.getPredicate().compareTo(key); 
			
			if (cmp < 0) {
				low = mid + 1;
			} else if (cmp > 0) {
				high = mid - 1;
			} else {
				while (mid > 0 && a.get(mid - 1).atom.getPredicate().equals(key)) {
					mid--;
				}
				
				return mid;
			}
		}
		
		return -(low + 1);
	}

    public static int binarySearch(Atom[] a, Predicate key) {
		int low = 0;
		int high = a.length - 1;
		
		while (low <= high) {
			int mid = (low + high) >>> 1;
			Atom midVal = a[mid];
			
			int cmp = midVal.getPredicate().compareTo(key); 
			
			if (cmp < 0) {
				low = mid + 1;
			} else if (cmp > 0) {
				high = mid - 1;
			} else {
				while (mid > 0 && a[mid - 1].getPredicate().equals(key)) {
					mid--;
				}
				
				return mid;
			}
		}
		
		return -(low + 1);
	}
}
