package edu.ntua.ivml.qa.lp;

import java.util.HashSet;
import java.util.Set;

public class Atom implements Cloneable, Comparable<Atom> {
	
	protected Predicate pred;
	protected Tuple args;
	
	public Atom(Predicate pred, Term... args) {
		this(pred, new Tuple(args));
	}
	
	public Atom(Predicate pred, Tuple args) {
		this.pred = pred;
		this.args = args;
	}
	
	public boolean isRole() {
		return getPredicate().isRole();
	}

	public boolean isConcept() {
		return getPredicate().isConcept();
	}
	
	public Predicate getPredicate() {
		return pred;
	}
	
	public Tuple getArguments() {
		return args;
	}	

	public Term getArgument(int i) {
		return args.getTerm(i);
	}	
	
	public Atom apply(Substitution subst) {
		return new Atom(pred, args.apply(subst));
	}

	public Set<Variable> getVariables() {
		return 	args.getVariables();
	}	
	

	public Set<Term> getTerms() {
		Set<Term> res = new HashSet<Term>();
		
		for (int i = 0; i < args.size(); i++) {
			res.add(args.getTerm(i));
		}
		
		return res;
	}
	
	public int depth() {
    	return args.depth();
    }
	
	public Object clone() {
		return new Atom(pred, (Tuple)args.clone());
	}

	public int compareTo(Atom a) {
		int nc = pred.compareTo(a.getPredicate());
		if (nc != 0) {
			return nc;
		} else {
			return getArguments().compareTo(a.getArguments());
		}
    }
	
	public boolean isFunctionFree() {
		for (int i = 0; i < args.size(); i++) {
			Term t = args.getTerm(i);
			
			if (t.isFunctionalTerm()) {
				return false;
			}
		}
		
		return true;
	}
	
	public boolean equals(Object obj) {
		if (!(obj instanceof Atom)) {
			return false;
		} else {
			return pred.equals(((Atom)obj).getPredicate()) && args.equals(((Atom)obj).getArguments());
		}
	}
	
	public boolean unboundEquals(Object obj) {
		if (!(obj instanceof Atom)) {
			return false;
		} else {
			return pred.equals(((Atom)obj).getPredicate()) && args.unboundEquals(((Atom)obj).getArguments());
		}
	}		
	
	public int hashCode() {
		if (pred != null && args != null) {
			return pred.hashCode() + args.hashCode();
		} else {
			return 0;
		}
	}
		
	public String toString() {
		return pred.toString() + args.toString();
	}

	public boolean hasUnboundVars() {
		return args.hasUnboundVars();
	}
	
    public String getStringPrefix() {
    	
    	if (args.size() == 0) {
    		return pred.getName();
    	} else {
    		String arguments = "";
    		for(int i = 0; i < args.size(); i++){
    			arguments += args.getTerm(i).getStringPrefix();
    		}
    		
    		return pred.getName() + arguments;
    	}
    }	

}