package edu.ntua.ivml.qa.lp;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

public class Constant extends SimpleTerm {

	public Constant(String name) {
		super(name);
	}

	public Set<Variable> getVariables() {
		return new HashSet<Variable>();
	}

	public ArrayList<Variable> getOrderedVariables() {
		return new ArrayList<Variable>();
	}	

	public boolean isConstant() {
		return true;
	}	
	
	public boolean isVariable() {
		return false;
	}
	
	public boolean isFunctionalTerm() {
		return false;
	}	
	
	public boolean hasUnboundVars() {
		return false;
	}	
	
    public int depth() {
    	return 0;
    }
    
	public Term apply(Substitution subst) {
		return this;
	}

	public Object clone() {
		Constant c = (Constant)super.clone();
		c.name = name;
		return c;
	}		
	
	public int compareTo(Term t) {
		if (t instanceof Variable) {
			return 1;
		} else if (t instanceof FunctionalTerm) {
			return -1;
		}
		
		return name.compareTo(((Constant)t).getName());
	}
	
	public boolean equals(Object obj) {
		if (!(obj instanceof Constant)) {
			return false;
		}
		
		return name.equals(((Constant)obj).getName());
	}
	
	public int hashCode() {
		return name.hashCode();
	}
	
	public String toString() {
		return name;
	}
	
	public String getStringPrefix() {
		return name;
	}

	
}
