package edu.ntua.ivml.qa.lp;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class LogicProgram {

	protected ArrayList<Clause> rules;
	protected Map<Predicate, ArrayList<Clause>> ruleIndex;
	
	public LogicProgram() {
		rules = new ArrayList<Clause>();
		ruleIndex = new HashMap<Predicate, ArrayList<Clause>>();
	}
	
	public LogicProgram(Collection<Clause> set) {		
		this();
		
		for (Clause c : set) {
			addClause(c);
		}
	}
	
	public int size() {
		return rules.size();
	}
	
	public void addClause(Clause r) {
		
		if (!rules.contains(r)) {
			rules.add(r);
			
			ArrayList<Clause> clauses = ruleIndex.get(r.getHead().getPredicate());
			if (clauses == null) {
				clauses = new ArrayList<Clause>();
				ruleIndex.put(r.getHead().getPredicate(), clauses);
			}
			clauses.add(r);
		}
	}
	
	public ArrayList<Clause> getClauses() {
		return rules;
	}
	
	public ArrayList<Clause> getClauses(Predicate pred) {
		ArrayList<Clause> res = ruleIndex.get(pred);
		if (res == null) {
			return new ArrayList<Clause>();
		} else {
			return res;
		}
	}	

	public Set<Predicate> getClausePredicates() {
		return ruleIndex.keySet();
	}
	
	public ArrayList<Clause> getGraphDependencyProgram(Predicate pred) {
		Set<Predicate> preds = new HashSet<Predicate>();
		preds.add(pred);
		
		return getGraphDependencyProgram(rules, preds);
	}
	
	public ArrayList<Clause> getGraphDependencyProgram(Set<Predicate> preds) {
		return getGraphDependencyProgram(rules, preds);
	}	
	
	public static ArrayList<Clause> getGraphDependencyProgram(ArrayList<Clause> clauses, Predicate pred) {
		Set<Predicate> preds = new HashSet<Predicate>();
		preds.add(pred);
		
		return getGraphDependencyProgram(clauses, preds);
	}
	
	public static ArrayList<Clause> getGraphDependencyProgram(Collection<Clause> clauses, Set<Predicate> pred) {
		
		Set<Predicate> predicatesToProcess = new HashSet<Predicate>();
		Set<Predicate> reachablePredicates = new HashSet<Predicate>();
		
		predicatesToProcess.addAll(pred);
		while(predicatesToProcess.size() > 0){
			Predicate p = predicatesToProcess.iterator().next();
			predicatesToProcess.remove(p);
			
			if(!reachablePredicates.contains(p)){
				reachablePredicates.add(p);
				predicatesToProcess.addAll(getDirectReachablePredicates(clauses, p));
			}
		}
		
		ArrayList<Clause> result = new ArrayList<Clause>();
		
		for(Clause r : clauses) {
			if(reachablePredicates.contains(r.getHead().getPredicate())) { 
				result.add(r);
			}
		}
		
		return result;
	}	
	
	
	private static Set<Predicate> getDirectReachablePredicates(Collection<Clause> clauses, Predicate p) {
		Set<Predicate> res = new HashSet<Predicate>();
		
		for (Clause r : clauses) {
			if (p.equals(r.getHead().getPredicate())) {
				for (Atom t : r.getBody()) {
					res.add(t.getPredicate());
				}
			}
		}
		return res;
	}
	

	public String toString() {
		String  s = "";
		
		for (int i = 0; i < rules.size(); i++) {
			s += i + ": " + rules.get(i).toString() + "\n";
		}
		return s;
	}
	
}
